
#ifndef PLS_TBB_LIKE_TASK_H
#define PLS_TBB_LIKE_TASK_H

#include "pls/internal/base/aligned_stack.h"
#include "pls/internal/base/deque.h"

#include "abstract_task.h"
#include "thread_state.h"

namespace pls {
    namespace internal {
        namespace scheduling {
            class fork_join_task;
            class fork_join_sub_task: public base::deque_item {
                friend class fork_join_task;

                // Coordinate finishing of sub_tasks
                std::atomic_uint32_t ref_count_;
                fork_join_sub_task* parent_;

                // Access to TBB scheduling environment
                fork_join_task* tbb_task_;

                // Stack Management (reset stack pointer after wait_for_all() calls)
                base::aligned_stack::state stack_state_;
            protected:
                explicit fork_join_sub_task();
                fork_join_sub_task(const fork_join_sub_task& other);

                virtual void execute_internal() = 0;
                // SubClass Implementations:
                // Do Work
                // |-- Spawn Sub Task (new subtask; spawn(subtask);)
                // |-- Spawn Sub task
                // Do Work
                // |-- Wait For All
                // Do  Work
                // |-- Spawn Sub Task

                template<typename T>
                void spawn_child(const T& sub_task);
                void wait_for_all();
            private:
                void spawn_child_internal(fork_join_sub_task* sub_task);
                void execute();

            public:
                virtual void test() {
                    std::cout << "Test" << std::endl;
                }
            };

            class fork_join_task: public abstract_task {
                friend class fork_join_sub_task;

                fork_join_sub_task* root_task_;
                base::aligned_stack* my_stack_;

                // Double-Ended Queue management
                base::deque<fork_join_sub_task> deque_;

                // Steal Management
                fork_join_sub_task* last_stolen_;

                fork_join_sub_task* get_local_sub_task();
                fork_join_sub_task* get_stolen_sub_task();

                bool internal_stealing(abstract_task* other_task) override;
                bool split_task(base::spin_lock* /*lock*/) override;

            public:
                explicit fork_join_task(fork_join_sub_task* root_task, const abstract_task::id& id):
                        abstract_task{0, id},
                        root_task_{root_task},
                        my_stack_{nullptr},
                        deque_{},
                        last_stolen_{nullptr} {};

                void execute() override {
                    // Bind this instance to our OS thread
                    my_stack_ = base::this_thread::state<thread_state>()->task_stack_;
                    root_task_->tbb_task_ = this;
                    root_task_->stack_state_ = my_stack_->save_state();

                    // Execute it on our OS thread until its finished
                    root_task_->execute();
                }
            };

            template<typename T>
            void fork_join_sub_task::spawn_child(const T& task)  {
                static_assert(std::is_base_of<fork_join_sub_task, T>::value, "Only pass fork_join_sub_task subclasses!");

                T* new_task = tbb_task_->my_stack_->push(task);
                spawn_child_internal(new_task);
            }
        }
    }
}

#endif //PLS_TBB_LIKE_TASK_H
